"""timecast.utils.ar"""
from typing import Any
from typing import Tuple

import jax
import jax.numpy as jnp
import numpy as np

from timecast.utils import internalize
from timecast.utils.gram import OnlineGram


def historify(
    X: np.ndarray,
    history_len: int,
    num_histories: int = None,
    offset: int = 0,
    # TODO: Remove this flag once we get rid of AR/PCR
    flatten: bool = False,
):
    """Generate (num_histories, history_len, input_dim) history from time series data

    Todo:
        * Implement striding

    Args:
        X (np.ndarray): first axis is time, remainder are feature dimensions
        history_len: length of a history window
        num_histories: number of history windows to create. Default is the max
        number of histories possible
        offset: how many time steps to offset
        flatten: flatten history window

    Returns:
        np.ndarray: nD array organized as (num_histories, history_len) + feature_shape
    """
    num_histories = (
        num_histories if num_histories is not None else (X.shape[0] - offset - history_len + 1)
    )

    if num_histories < 1 or history_len < 1:
        raise ValueError("Must have positive history_len and at least one window")
    if X.shape[0] < offset + num_histories + history_len - 1:
        raise ValueError(
            "Not enough history ({}) to produce {} windows of length {} with offset {}".format(
                X.shape[0], num_histories, history_len, offset
            )
        )

    shape = (num_histories, -1)
    if not flatten:
        shape = (num_histories, history_len) + X.shape[1:]

    return jnp.swapaxes(
        jnp.stack([jnp.roll(X, shift=-(i + offset), axis=0) for i in range(history_len)]), 0, 1
    )[:num_histories].reshape(shape)


def _form_constraints(
    input_dim: int, output_dim: int, history_len: int, fit_intercept: bool = True
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """
    Description: Sets the linear constraint matrix and vector R and r,
    respectively for constraining the `_params` vector. Specifically, each row
    in R is a linearly independent constraint where we force each dimension in
    the time series value to share a coefficient across the window.

    References: https://www.le.ac.uk/users/dsgp1/COURSES/TOPICS/restrict.pdf

    Args:
        input_dim (int): number of dimensions in the time series value
        output_dim (int): number of dimensions in the output
        history_len (int): window size for the auto reggressor
        fit_intercept (bool, optional): whether or not to fit an intercept.
        Defaults to True.

    Returns:
        Tuple[np.ndarray, np.ndarray]: RB = r
    """
    if fit_intercept:
        history_len += 1

    num_constraints = history_len * (input_dim - 1)

    R = np.zeros((num_constraints, input_dim * history_len))
    r = np.zeros((num_constraints, output_dim))

    # To share a parameter p across features x1 and x2, create a new row in R
    # with index 1 set to 1, index 2 set to -1, the rest set to 0 and the
    # corresponding row in r set to 0 (i.e., x1 + x2 = 0)
    for i in range(history_len):
        for j in range(input_dim - 1):

            # Set the constraint row
            row = i * (input_dim - 1) + j

            # Column indices for the first and second feature to tie. Note that
            # within the inner for loop, we tie multiple features to the same
            # parameter
            col1 = i * input_dim + j
            col2 = i * input_dim + j + 1

            # Update R appropriately
            R[row, col1] = 1
            R[row, col2] = -1

    return R, r


def _fit_constrained(beta, inv, R, r):
    """Fit constrained"""
    return beta - inv @ (R.T @ np.linalg.inv(R @ (inv @ R.T))) @ (R @ beta - r)


def _compute_xtx_inverse(XTX, alpha):
    """Compute inverse of X.T @ X"""
    reg = alpha * jnp.eye(XTX.shape[0])
    reg = jax.ops.index_update(reg, [0, 0], 0)
    inv = jnp.linalg.inv(XTX + reg)
    return inv


def _fit_unconstrained(inv, XTY):
    """Fit unconstrained"""
    return inv @ XTY


def fit_gram(
    XTX: OnlineGram,
    XTY: OnlineGram,
    alpha: float = 1.0,
    normalize: bool = False,
    projection: np.ndarray = None,
    input_dim: int = None,
):
    """Compute linear regression parameters from gram matrix

    Notes:
        * Assumes over-determined systems
        * Assumes we always fit an intercept
    """

    fit_intercept = True
    feature_dim = XTX.feature_dim if projection is None else projection.shape[1]
    output_dim = XTY.output_dim

    if input_dim is None:
        history_len = None
    else:
        if feature_dim % input_dim != 0:
            raise ValueError("Original input dimension must evenly divide feature dimensions")
        history_len = feature_dim // input_dim

    if XTX.observations <= feature_dim:
        raise ValueError(
            "Underdetermined systems not currently supported (observations: {},"
            "features: {})".format(XTX.observations, feature_dim)
        )

    # Finalize gram matrices
    XTX = XTX.matrix(
        normalize=normalize,
        projection=projection,
        fit_intercept=fit_intercept,
        input_dim=input_dim,
    )
    XTY = XTY.matrix(
        normalize=normalize,
        projection=projection,
        fit_intercept=fit_intercept,
        input_dim=input_dim,
    )
    inv = _compute_xtx_inverse(XTX, alpha)
    beta = _fit_unconstrained(inv, XTY)

    # Ignore constrained regression if we project
    if input_dim is not None:
        R, r = _form_constraints(
            input_dim=input_dim,
            output_dim=output_dim,
            history_len=history_len,
            fit_intercept=fit_intercept,
        )
        beta = _fit_constrained(beta, inv, R, r)
        beta = beta.take(jnp.arange(0, len(beta), input_dim), axis=0)
        return beta[1:], beta[0]

    return beta[1:].reshape(1, feature_dim, -1), beta[0]


def compute_gram(
    data: Tuple[np.ndarray, np.ndarray, Any], input_dim: int, output_dim: int, history_len: int
) -> Tuple[OnlineGram, OnlineGram]:
    """Compute X.T @ X and X.T @ Y on history windows incrementally"""
    num_features = input_dim * history_len
    XTX = OnlineGram(num_features)
    XTY = OnlineGram(num_features, output_dim)

    for X, Y, _ in data:
        X = internalize(X, input_dim)[0]
        Y = internalize(Y, output_dim)[0]

        if X.shape[0] != Y.shape[0]:
            raise ValueError("Input and output data must have the same number of observations")

        # Expand input time series X into histories, whic should result in a
        # (num_histories, history_len * input_dim)-shaped array
        history = historify(X, history_len=history_len, flatten=True)

        XTX.update(history)
        XTY.update(history, Y[history_len - 1 :])

    if XTX.observations == 0:
        raise IndexError("No data to fit")

    if XTX.observations <= num_features:
        raise ValueError(
            "Underdetermined systems not currently supported (observations: {},"
            "features: {})".format(XTX.observations, num_features)
        )

    return XTX, XTY



"""Linear transformation"""
from typing import Iterable

import flax
import jax
import jax.numpy as jnp

from timecast.learners.base import NewMixin

default_kernel_init = flax.nn.initializers.lecun_normal()


def _normalize_axes(axes, ndim):
    """
    A tuple by convention. len(axes_tuple) then also gives the rank efficiently.

    Warning:
        * This doesn't raise when some axis is greater than ndim - 1
    """
    return tuple([ax if ax >= 0 else ndim + ax for ax in axes])


class Linear(NewMixin, flax.nn.Module):
    """A linear transformation with flexible axes."""

    def apply(
        self,
        inputs,
        output_shape,
        input_axes=-1,
        batch_axes=(),
        bias=True,
        dtype=jnp.float32,
        kernel_init=default_kernel_init,
        bias_init=flax.nn.initializers.zeros,
        precision=None,
    ):
        """Applies a linear transformation to the inputs along multiple dimensions.
    Args:
      inputs: The nd-array to be transformed.
      output_shape: tuple of output shape.
      input_axes: tuple with axes to apply the transformation on.
      batch_axes: tuple with batch axes.
      bias: whether to add a bias to the output (default: True).
      dtype: the dtype of the computation (default: float32).
      kernel_init: initializer function for the weight matrix.
      bias_init: initializer function for the bias.
      precision: numerical precision of the computation see `jax.lax.Precision`
        for details.
    Returns:
      The transformed input.
    """
        inputs = jnp.asarray(inputs, dtype)

        if not isinstance(output_shape, Iterable):
            output_shape = (output_shape,)
        if not isinstance(input_axes, Iterable):
            input_axes = (input_axes,)
        if not isinstance(batch_axes, Iterable):
            batch_axes = (batch_axes,)
        output_shape, input_axes, batch_axes = (
            tuple(output_shape),
            tuple(input_axes),
            tuple(batch_axes),
        )

        if len(input_axes) == 0:
            raise IndexError("Must have at least one input dimension")

        if len(output_shape) == 0:
            raise IndexError("Must have at least one output dimension")

        if batch_axes:
            max_dim = jnp.max(batch_axes)
            if set(batch_axes) != set(range(max_dim + 1)):
                raise ValueError(
                    "batch_axes %s must be consecutive leading "
                    "dimensions starting from 0." % str(batch_axes)
                )

        if set(batch_axes) & set(input_axes):
            raise IndexError("Batch axes and input axes must not have reused axes")

        if jnp.max(batch_axes + input_axes) >= inputs.ndim:
            raise IndexError("Not enough dimensions in input for batch/input axes")

        ndim = inputs.ndim
        input_axes = _normalize_axes(input_axes, ndim)
        batch_axes = _normalize_axes(batch_axes, ndim)
        n_input_axes, n_output_axes = len(input_axes), len(output_shape)

        def kernel_init_wrap(rng, shape, dtype=jnp.float32):
            """Initializing and inducing correct shapes"""
            flat_shape = (
                jnp.prod(shape[:n_input_axes]),
                jnp.prod(shape[-n_output_axes:]),
            )
            kernel = kernel_init(rng, flat_shape, dtype)
            return jnp.reshape(kernel, shape)

        kernel_shape = tuple([inputs.shape[ax] for ax in input_axes]) + output_shape
        kernel = self.param("kernel", kernel_shape, kernel_init_wrap)
        kernel = jnp.asarray(kernel, dtype)

        contract_ind = tuple(range(n_input_axes))
        out = jax.lax.dot_general(
            inputs, kernel, ((input_axes, contract_ind), ((), ())), precision=precision
        )

        if bias:

            def bias_init_wrap(rng, shape, dtype=jnp.float32):
                """Initializing and inducing correct shapes"""
                flat_shape = (jnp.prod(shape[-n_output_axes:]),)
                bias = bias_init(rng, flat_shape, dtype)
                return jnp.reshape(bias, shape)

            bias = self.param("bias", output_shape, bias_init_wrap)
            # Reshape bias for broadcast.
            num_expand_dims = len(set(range(inputs.ndim)) - set(input_axes))
            for _ in range(num_expand_dims):
                bias = bias[None, :]
            bias = jnp.asarray(bias, dtype)
            out = out + bias

        return out



"timecast.series.random"
from typing import Tuple

import jax.numpy as jnp
import numpy as np


def generate(n: int = 1000, loc: float = 0.0, scale: float = 1.0) -> Tuple[np.ndarray, np.ndarray]:
    """
    Description: outputs a timeline randomly distributed i.i.d. from gaussian
    with mean `loc`, standard deviation `scale`
    """
    X = np.random.normal(loc=loc, scale=scale, size=(n + 1))

    return jnp.asarray(X)[:-1].reshape(-1, 1), jnp.asarray(X)[1:].reshape(-1, 1)



"""flax.nn.Module for an auto-regressive online learner.

Todo:
    * Implement batching
    * Implement projections
"""
from typing import Tuple
from typing import Union

import flax
import jax
import jax.numpy as jnp
import numpy as np

from timecast.learners._arx_history import ARXHistory
from timecast.learners.base import NewMixin
from timecast.utils.ar import historify


def _none_init(rng, shape):
    """Initialize with scalar None"""
    return None


class ARX(NewMixin, flax.nn.Module):
    """AR online learner"""

    def apply(
        self,
        targets: np.ndarray = None,
        features: np.ndarray = None,
        history_len: int = 1,
        output_shape: Union[Tuple[int, ...], int] = 1,
        constrain: bool = True,
    ):
        """
        Notation
            * x = features
            * y = targets
            * H = history_len

        Estimates the following:
        
            \hat{y} = \sum_{i = 1}^{H + 1} B_i x_{t - i - 1} + a
                      \sum_{i = 1} ^ H A_i y_{t - i} + b

        Notes:
            * Assumes `features` and `targets` are the shape of one time step
            of data
            * Delegates much of the error checking to ARXHistory

        Args:
            targets (np.ndarray): target data
            features (np.ndarray): feature data
            output_shape (Union[Tuple[int, ...], int]): int or tuple
            describing output shape
            history_len (int): length of history
            constrain: force one parameter per for each slot in history. TODO:
            explain this better

        Returns:
            np.ndarray: result
        """

        # TODO: Check and reshape inputs; for now assumes data is time x
        # dimension
        # - We have some shape checks, but not at all robust

        if history_len < 1:
            raise ValueError("Features require a history length of at least 1")

        has_targets = targets is not None and targets.ndim > 0
        has_features = features is not None and features.ndim > 0

        if has_targets and targets.ndim == 1:
            targets = targets.reshape(1, -1)

        if has_features and features.ndim == 1:
            features = features.reshape(1, -1)

        self.T = self.state("T", shape=())
        self.target_history = self.state("target_history", shape=(), initializer=_none_init)
        self.feature_history = self.state("feature_history", shape=(), initializer=_none_init)

        if self.is_initializing():
            self.T.value = 0

            if has_targets:
                target_history_shape = (history_len, targets.shape[1])
                self.target_history.value = jnp.zeros(target_history_shape)
            if has_features:
                feature_history_shape = (history_len, features.shape[1])
                self.feature_history.value = jnp.zeros(feature_history_shape)

        target_histories, feature_histories = None, None
        if has_targets:
            target_histories = historify(
                jnp.vstack((self.target_history.value, targets))[:-1, :], history_len=history_len
            )

        if has_features:
            feature_histories = historify(
                jnp.vstack((self.feature_history.value, features))[1:, :], history_len=history_len
            )

        y_hat = ARXHistory(
            targets=target_histories,
            features=feature_histories,
            output_shape=output_shape,
            constrain=constrain,
        )

        # TODO: Don't duplicate the vstacks (modulo index difference for target_history)
        if not self.is_initializing():
            # Update target history with data _after_ we have made calculations
            if has_targets:
                self.target_history.value = jnp.vstack((self.target_history.value, targets))[
                    targets.shape[0] :
                ]
            if has_features:
                self.feature_history.value = jnp.vstack((self.feature_history.value, features))[
                    features.shape[0] :
                ]

            self.T.value += 1

        # If we have targets, then we need to wait one additional time step to
        # have a full target window
        return jax.lax.cond(
            self.T.value + (1 if has_targets else 0) >= history_len,
            y_hat,
            lambda x: x,
            y_hat,
            lambda x: jax.lax.stop_gradient(y_hat),
        )
        # if self.T.value + (1 if has_targets else 0) >= history_len:
        # return y_hat
        # else:
        # return jax.lax.stop_gradient(y_hat)



"""timecast.optim

Todo:
    * Document available optimizers
"""
import flax
from flax.optim import Adagrad
from flax.optim import Adam
from flax.optim import GradientDescent
from flax.optim import LAMB
from flax.optim import LARS
from flax.optim import Momentum
from flax.optim import RMSProp


@flax.struct.dataclass
class _DummyGradHyperParams:
    """DummyGrad hyper parameters"""

    add: float


class DummyGrad(flax.optim.OptimizerDef):
    """Dummy optimizer for testing"""

    def __init__(self, add: float = 0.0):
        """Initialize hyper parameters"""
        super().__init__(_DummyGradHyperParams(add))

    def init_param_state(self, param):
        """Initialize parameter state"""
        return {}

    def apply_param_gradient(self, step, hyper_params, param, state, grad):
        """Apply per-parametmer gradients"""
        new_param = param + hyper_params.add
        return new_param, state


__all__ = ["Adagrad", "Adam", "GradientDescent", "Momentum", "LAMB", "LARS", "RMSProp"]



"""flax.nn.Module for wrapping an array

Todo:
    * Validate data and add negative tests
    * Implement batching
"""
import flax


class Precomputed(flax.nn.Module):
    """Wraps an array

    Notes:
        * Assumes the first dimension is a time dimension
        * Assumes the data is accessed in order (i.e., no shuffling)
    
    Warning:
        * Ignores init_by_shape
    """

    def apply(self, x, arr):
        """Apply function"""
        self.index = self.state("index", shape=(), initializer=flax.nn.initializers.zeros)
        val = arr[self.index.value.astype(int)]
        if not self.is_initializing():
            self.index.value += 1
        return val



"""flax.nn.Module for an auto-regressive online learner.

Todo:
    * Implement batching; right now, passing in a batch skips to the end and
    predicts one value, rather than predict one value per example or some sort
    of averaging
    * Implement strided history
    * Add link functions for GLM

References:
    * http://eeweb.poly.edu/iselesni/lecture_notes/least_squares/least_squares_SP.pdf
"""
from typing import Any
from typing import Iterable
from typing import Tuple
from typing import Union

import flax
import jax.numpy as jnp
import numpy as np

from timecast.learners._linear import Linear
from timecast.learners.base import FitMixin
from timecast.learners.base import NewMixin
from timecast.utils import random
from timecast.utils.ar import compute_gram
from timecast.utils.ar import fit_gram


class AR(NewMixin, FitMixin, flax.nn.Module):
    """AR online learner"""

    def apply(
        self,
        x: np.ndarray,
        history_len: int,
        output_dim: Union[Tuple[int, ...], int] = 1,
        history: np.ndarray = None,
        loc: Union[np.ndarray, float] = None,
        scale: Union[np.ndarray, float] = None,
    ):
        """
        Note:
            * We expect that `x` is one- or two-dimensional
            * We reshape `x` to ensure its first axis is time and its second
              axis is input_features

        Args:
            x (np.ndarray): input data
            history_len (int): length of AR history length
            output_dim (Union[Tuple[int, ...], int]): int or tuple
            describing output shape
            history (np.ndarray, optional): Defaults to None. Optional
            initialization for history
            loc: mean for centering data
            scale: std for normalizing data

        Returns:
            np.ndarray: result
        """

        if jnp.isscalar(x):
            x = jnp.array([[x]])
        if x.ndim == 1:
            x = x.reshape(1, -1)

        self.history = self.state(
            "history", shape=(history_len, x.shape[1]), initializer=flax.nn.initializers.zeros
        )

        if self.is_initializing() and history is not None:
            self.history.value = jnp.vstack((self.history.value, history))[history.shape[0] :]
        elif not self.is_initializing():
            self.history.value = jnp.vstack((self.history.value, x))[x.shape[0] :]

        inputs = self.history.value.ravel().reshape(1, -1)

        if loc is not None:
            inputs -= loc

        if scale is not None:
            inputs /= scale

        y = Linear(
            inputs=inputs,
            output_shape=output_dim,
            input_axes=(0, 1),
            bias=True,
            dtype=jnp.float32,
            kernel_init=flax.nn.initializers.zeros,
            bias_init=flax.nn.initializers.zeros,
            precision=None,
            name="Linear",
        )
        return y

    @classmethod
    def fit(
        cls,
        data: Iterable[Tuple[np.ndarray, np.ndarray, Any]],
        input_dim: int,
        history_len: int,
        output_dim: int = 1,
        normalize: bool = True,
        alpha: float = 1.0,
        key: jnp.ndarray = None,
        history: np.ndarray = None,
        name: str = "AR",
        **kwargs
    ) -> flax.nn.Model:
        """Receives data as an iterable of tuples containing input time series,
        true time series

        Todo:
            * We assume input_dim is one-dimensional; we should flatten if not
            * Really intended for passing in timeseries at a time, not
            individual time series observations; is this the right general API?
            * Shape is (1, input_dim); what about mini-batches?

        Notes:
            * Use (1, history_len * input_dim) vectors as features (could
            consider other representations)
            * Given a time series of length N and a history of length H,
            construct N - H + 1 windows
            * We could infer input_dim from data, but for now, require
            users to explicitly provide
            * Assumes we get tuples of time series, not individual time series
            observations

        Args:
            data: an iterable of tuples containing input/truth pairs of time
            series plus any auxiliary value
            input_dim: number of feature dimensions in input
            history_len: length of history to consider
            output_dim: number of feature dimensions in output
            normalize: zscore data or not
            alpha: for ridge regression
            key: random key for jax random
            history: Any history to pass to AR
            name: name for the top-level module
            kwargs: Extra keyword arguments

        Returns:
            flax.nn.Model: initialized model
        """
        XTX, XTY = compute_gram(data, input_dim, output_dim, history_len)

        kernel, bias = fit_gram(XTX, XTY, normalize=normalize, alpha=alpha)

        loc = XTX.mean if normalize else None
        scale = XTX.std if normalize else None

        model_def = AR.partial(
            history_len=history_len,
            output_dim=output_dim,
            loc=loc,
            scale=scale,
            history=history,
            name=name,
        )

        if key is None:
            key = random.generate_key()

        with flax.nn.stateful() as state:
            _, params = model_def.init_by_shape(key, [(1, input_dim)])

        model = flax.nn.Model(model_def, params)
        model.params["Linear"]["kernel"] = kernel
        model.params["Linear"]["bias"] = bias

        return model, state



"""flax.nn.Module for an auto-regressive online learner.

Todo:
    * Implement strided histories
    * Figure out normalizing
"""
import warnings
from typing import Tuple
from typing import Union

import flax
import jax.numpy as jnp
import numpy as np

from timecast.learners._linear import Linear
from timecast.learners.base import NewMixin


default_output_shape: int = 1


class _ARXHistory(NewMixin, flax.nn.Module):
    """AR online learner helper"""

    def apply(
        self,
        data: np.ndarray,
        output_shape: Union[Tuple[int, ...], int] = default_output_shape,
        constrain: bool = True,
        name: str = "ARXHistory",
    ):
        """
        Args:
            data (np.ndarray): (batch, history_len, input_dim)
            output_shape (Union[Tuple[int, ...], int]): int or tuple
            describing output shape
            constrain: force one parameter per for each slot in history. TODO:
            explain this better
            name (str): name to pass to Linear

        Returns:
            np.ndarray: result
        """
        # TODO: Check shape of features and reshape if necessary
        input_shape = data.shape[2:]
        history_shape = data.shape[1:]

        # TODO (flax): We really shouldn't need state to just have some local
        # variables set in is_initializing...
        self.history_shape = self.state("history_shape", shape=(history_shape))
        if self.is_initializing():
            self.history_shape.value = history_shape

        # TODO: We would ideally like to be able to raise with jit
        """
        else:
            if history_shape != self.history_shape.value:
                raise ValueError(
                    "Got input_shape {}, expected input_shape {}".format(
                        history_shape, self.history_shape.value
                    )
                )
        """

        if jnp.isscalar(output_shape):
            output_shape = (output_shape,)

        if constrain:
            # If we have the non-default output_shape and it doesn't match input_shape
            if output_shape != default_output_shape and input_shape != output_shape:
                warnings.warn(
                    "When constrained, input data shape must equal output data"
                    "shape. Got input_shape {} and output_shape {}. Coercing"
                    "output_shape to input_shape".format(input_shape, output_shape)
                )
            output_shape = input_shape
            input_axes = (1,)
        else:
            input_axes = tuple(range(1, data.ndim))

        return Linear(
            inputs=data,
            output_shape=output_shape,
            input_axes=input_axes,
            batch_axes=(0,),
            bias=True,
            dtype=jnp.float32,
            kernel_init=flax.nn.initializers.zeros,
            bias_init=flax.nn.initializers.zeros,
            precision=None,
            name="Linear",
        )


class ARXHistory(NewMixin, flax.nn.Module):
    """AR online learner with history as input"""

    def apply(
        self,
        targets: np.ndarray = None,
        features: np.ndarray = None,
        output_shape: Union[Tuple[int, ...], int] = 1,
        constrain: bool = True,
    ):
        """
        Notation
            * x = features
            * y = targets
            * H = history_len

        Estimates the following:
        
            \hat{y} = \sum_{i = 1}^{H + 1} B_i x_{t - i - 1} + a
                      \sum_{i = 1} ^ H A_i y_{t - i} + b

        Notes:
            * Assumes `features` and `targets` have three dimensions: (batch,
            history, data). Any extra dimensions are part of the data shape
            * Doesn't care if features and targets have different history or
            input dimensions

        Args:
            targets (np.ndarray): target data
            features (np.ndarray): feature data
            output_shape (Union[Tuple[int, ...], int]): int or tuple
            describing output shape
            constrain: force one parameter per for each slot in history. TODO:
            explain this better

        Returns:
            np.ndarray: result
        """

        Ay, Bx = 0, 0

        self.has_targets = self.state("has_targets", shape=())
        self.has_features = self.state("has_features", shape=())

        has_targets = targets is not None and targets.ndim > 0
        has_features = features is not None and features.ndim > 0

        if self.is_initializing():
            self.has_targets.value = has_targets
            self.has_features.value = has_features
            # TODO: We would ideally like to be able to raise with jit
            """
            if not has_targets and not has_features:
                raise ValueError("Need one or both of targets and features")
            """

        # TODO: We would ideally like to be able to raise with jit
        """
        else:
            if not has_targets and self.has_targets.value:
                raise ValueError("Expected targets, but got None")
            if has_targets and not self.has_targets.value:
                raise ValueError("Did not expected targets, but got targets")
            if not has_features and self.has_features.value:
                raise ValueError("Expected features, but got None")
            if has_features and not self.has_features.value:
                raise ValueError("Did not expected features, but got features")
            if has_features and has_targets and features.shape[0] != targets.shape[0]:
                raise ValueError("Targets and features need to have same sized batch")
        """

        if has_targets:
            Ay = _ARXHistory(
                data=targets, output_shape=output_shape, constrain=constrain, name="Targets"
            )
        if has_features:
            Bx = _ARXHistory(
                data=features, output_shape=output_shape, constrain=constrain, name="Features"
            )

        return Ay + Bx



"""timecast.learners.precomputed: testing"""
import flax
import jax
import numpy as np

from timecast.learners import Precomputed
from timecast.utils import random


def test_precomputed():
    """Testing Precomputed"""
    arr = jax.random.uniform(random.generate_key(), shape=(10, 10))
    with flax.nn.stateful() as state:
        model_def = Precomputed.partial(arr=arr, name="Precomputed")
        _, params = model_def.init_by_shape(random.generate_key(), [(10, 10)])
        model = flax.nn.Model(model_def, params)

    for i in range(arr.shape[0]):
        with flax.nn.stateful(state) as state:
            np.testing.assert_array_equal(arr[i], model(None))



"""timecast.learners._arx_history: testing"""
import flax
import jax
import jax.numpy as jnp
import numpy as np
import pytest

from timecast.learners._arx_history import _ARXHistory
from timecast.utils import random


@pytest.mark.parametrize("constrain", [True, False])
@pytest.mark.parametrize("history_len", [1, 5])
@pytest.mark.parametrize("input_dim", [1, 5])
@pytest.mark.parametrize("batch_size", [1, 5])
def test_arx_history_helper(constrain, history_len, input_dim, batch_size):
    """Test ARX helper"""
    output_shape = input_dim
    model, state = _ARXHistory.new(
        shapes=[(1, history_len, input_dim)], output_shape=output_shape, constrain=constrain,
    )

    shape = (batch_size, history_len, input_dim)
    inputs = jax.random.uniform(random.generate_key(), shape=shape)

    kernel_shape = (
        (history_len, output_shape) if constrain else (history_len, input_dim, output_shape)
    )
    kernel = jax.random.uniform(random.generate_key(), shape=kernel_shape)
    bias = jax.random.uniform(random.generate_key(), shape=(output_shape,))

    model.params["Linear"]["kernel"] = kernel
    model.params["Linear"]["bias"] = bias

    with flax.nn.stateful(state) as state:
        result = model(inputs)

    inputs_axes = (1,) if constrain else (1, 2)
    num_expand_dims = len(set(range(inputs.ndim)) - set(inputs_axes))
    for _ in range(num_expand_dims):
        bias = bias[None, :]
    expected = (
        jnp.tensordot(inputs, kernel, axes=[inputs_axes, tuple(range(len(inputs_axes)))]) + bias
    )

    np.testing.assert_array_almost_equal(expected, result)


"""
def test_arx_history_helper_inconsistent_history_shape():
    model, state = _ARXHistory.new(shapes=[(1, 5, 10)], constrain=False)
    with flax.nn.stateful(state) as state:
        with pytest.raises(ValueError):
            model(jnp.ones((10, 1, 1)))


def test_arx_history_errors():
    targets = jnp.ones((10, 4, 1))
    features = jnp.ones((10, 2, 10))

    model, state = ARXHistory.new(shapes=[targets.shape, features.shape], constrain=False)
    model, state = ARXHistory.new(shapes=[targets.shape, features.shape], constrain=True)

    with pytest.raises(ValueError):
        model_def = ARXHistory.partial(constrain=False)
        with flax.nn.stateful() as _:
            _, params = model_def.init_by_shape(random.generate_key(), [])

    with pytest.raises(ValueError):
        model_def = ARXHistory.partial(constrain=False)
        with flax.nn.stateful() as _:
            _, params = model_def.init_by_shape(
                random.generate_key(), [targets.shape, features.shape]
            )
            model = flax.nn.Model(model_def, params)
            model(jnp.ones((5, 2, 10)), jnp.ones((4, 4, 1)))

    with pytest.raises(ValueError):
        model_def = ARXHistory.partial(constrain=False)
        with flax.nn.stateful() as _:
            _, params = model_def.init_by_shape(random.generate_key(), [(), features.shape])
            model = flax.nn.Model(model_def, params)
            model(targets, features)

    with pytest.raises(ValueError):
        model_def = ARXHistory.partial(constrain=False)
        with flax.nn.stateful() as _:
            _, params = model_def.init_by_shape(random.generate_key(), [targets.shape])
            model = flax.nn.Model(model_def, params)
            model(targets, features)

    with pytest.raises(ValueError):
        model_def = ARXHistory.partial(constrain=False)
        with flax.nn.stateful() as _:
            _, params = model_def.init_by_shape(
                random.generate_key(), [targets.shape, features.shape]
            )
            model = flax.nn.Model(model_def, params)
            model(features=features)

    with pytest.raises(ValueError):
        model_def = ARXHistory.partial(constrain=False)
        with flax.nn.stateful() as _:
            _, params = model_def.init_by_shape(
                random.generate_key(), [targets.shape, features.shape]
            )
            model = flax.nn.Model(model_def, params)
            model(targets)
"""



"""timecast.utils.ar: testing"""
import jax
import jax.numpy as jnp
import numpy as np
import pytest

from tests.learners.test_ar import _compute_kernel_bias
from timecast.utils import random
from timecast.utils.ar import _compute_xtx_inverse
from timecast.utils.ar import _fit_constrained
from timecast.utils.ar import _fit_unconstrained
from timecast.utils.ar import _form_constraints
from timecast.utils.ar import compute_gram
from timecast.utils.ar import fit_gram
from timecast.utils.ar import historify
from timecast.utils.gram import OnlineGram


@pytest.mark.parametrize("m", [1, 10])
@pytest.mark.parametrize("n", [1, 10])
@pytest.mark.parametrize("history_len", [-1, 0, 1, 10])
@pytest.mark.parametrize("num_histories", [0, 1, 10])
@pytest.mark.parametrize("offset", [0, 1, 10])
def test_historify(m, n, history_len, num_histories, offset):
    """Test history-making"""
    X = jax.random.uniform(random.generate_key(), shape=(m, n))

    if (
        num_histories < 1
        or history_len < 1
        or X.shape[0] < offset + num_histories + history_len - 1
    ):
        with pytest.raises(ValueError):
            historify(X, num_histories, history_len, offset)

    else:
        batched = historify(X, history_len, num_histories, offset)

        for i, batch in enumerate(batched):
            np.testing.assert_array_almost_equal(X[i + offset : i + offset + history_len], batch)


@pytest.mark.parametrize("n", [100])
@pytest.mark.parametrize("input_dim", [1, 5])
@pytest.mark.parametrize("output_dim", [1, 4])
@pytest.mark.parametrize("history_len", [1, 3])
def test_compute_gram(n, input_dim, output_dim, history_len):
    """Test compouting gram matrices"""
    X = jax.random.uniform(random.generate_key(), shape=(n, input_dim))
    Y = jax.random.uniform(random.generate_key(), shape=(n, output_dim))

    XTX, XTY = compute_gram([(X, Y, None)], input_dim, output_dim, history_len)

    history = historify(X, history_len, n - history_len + 1, flatten=True)
    np.testing.assert_array_almost_equal(history.T @ history, XTX.matrix(), decimal=4)
    np.testing.assert_array_almost_equal(history.T @ Y[history_len - 1 :], XTY.matrix(), decimal=4)


def test_compute_gram_no_data():
    """Test no data"""
    with pytest.raises(ValueError):
        compute_gram([(jnp.zeros((0, 1)), jnp.zeros((0, 1)), None)], 1, 1, 1)

    with pytest.raises(IndexError):
        compute_gram([], 1, 1, 1)


def test_compute_gram_underdetermined():
    """Test underdetermined"""
    data = jnp.ones((13, 10))
    with pytest.raises(ValueError):
        compute_gram([(data, data, None)], 10, 10, 10)


def test_fit_gram_underdetermined():
    """Test underdetermined"""
    XTX = OnlineGram(1)
    XTY = XTX

    with pytest.raises(ValueError):
        fit_gram(XTX, XTY)


@pytest.mark.parametrize(
    "history_len,input_dim,output_dim,fit_intercept,expected_R,expected_r",
    [
        (
            3,
            2,
            1,
            False,
            np.array(
                [
                    [1.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, -1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, -1.0],
                ]
            ),
            np.zeros(3),
        ),
        (
            3,
            2,
            1,
            True,
            np.array(
                [
                    [1.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, -1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, -1.0],
                ]
            ),
            np.zeros(4),
        ),
        (4, 1, 1, True, np.zeros((0, 5)), np.zeros((0))),
        (
            3,
            2,
            2,
            True,
            np.array(
                [
                    [1.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, -1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, -1.0],
                ]
            ),
            np.zeros((4, 2)),
        ),
    ],
)
def test_form_constraints(
    history_len, input_dim, output_dim, fit_intercept, expected_R, expected_r
):
    """Test forming constraints"""
    R, r = _form_constraints(input_dim, output_dim, history_len, fit_intercept)

    r = r.squeeze()

    assert np.array_equal(expected_R, R)
    assert np.array_equal(expected_r, r)


@pytest.mark.parametrize("n", [40, 1000])
@pytest.mark.parametrize("input_dim", [1, 5])
@pytest.mark.parametrize("output_dim", [1, 4])
@pytest.mark.parametrize("history_len", [1, 3])
def test_fit_unconstrained(n, input_dim, output_dim, history_len):
    """Fit unconstrained regression"""
    # NOTE: we use random data because we want to test dimensions and
    # correctness vs a second implementation
    X = jax.random.uniform(random.generate_key(), shape=(n, input_dim))
    Y = jax.random.uniform(random.generate_key(), shape=(n, output_dim))

    XTX, XTY = compute_gram([(X, Y, None)], input_dim, output_dim, history_len)

    kernel, bias = fit_gram(XTX, XTY)
    num_histories = n - history_len + 1
    history = historify(X, history_len, num_histories, flatten=True)

    expected_kernel, expected_bias = _compute_kernel_bias(history, Y[history_len - 1 :], alpha=1.0)
    expected_kernel = expected_kernel.reshape(1, history_len * input_dim, -1)

    np.testing.assert_array_almost_equal(expected_kernel, kernel, decimal=3)
    np.testing.assert_array_almost_equal(expected_bias, bias, decimal=3)


@pytest.mark.parametrize("n", [1000])
@pytest.mark.parametrize("input_dim", [10, 12])
@pytest.mark.parametrize("output_dim", [1, 10])
@pytest.mark.parametrize("history_len", [2])
def test_fit_constrained(n, input_dim, output_dim, history_len):
    """Fit constrained regression"""
    # NOTE: we use random data because we want to test dimensions and
    # correctness vs a second implementation
    X = jax.random.uniform(random.generate_key(), shape=(n, input_dim))
    Y = jax.random.uniform(random.generate_key(), shape=(n, output_dim))

    XTX, XTY = compute_gram([(X, Y, None)], input_dim, output_dim, history_len)
    result = fit_gram(XTX, XTY, input_dim=input_dim)

    # Next, check that each chunk of input_dim features have the same coefficient
    # result = fit_gram(XTX, XTY, input_dim=input_dim)
    R, r = _form_constraints(
        input_dim=input_dim, output_dim=output_dim, history_len=history_len, fit_intercept=True,
    )

    XTX = XTX.matrix(fit_intercept=True, input_dim=input_dim)
    XTY = XTY.matrix(fit_intercept=True, input_dim=input_dim)
    inv = _compute_xtx_inverse(XTX, alpha=1.0)
    beta = _fit_unconstrained(inv, XTY)
    beta = _fit_constrained(beta, inv, R, r)
    beta = beta.reshape(history_len + 1, input_dim, -1)
    assert np.sum([np.abs(x - x[0]) for x in beta]) < 1e-4

    # Finally, check that resulting vector is of the correct length and the
    # values are self-consistent
    assert len(beta) == history_len + 1

    beta = beta[:, 0]
    beta = beta[1:], beta[0]

    # Check final results
    np.testing.assert_array_almost_equal(beta[0], result[0])
    np.testing.assert_array_almost_equal(beta[1], result[1])


def test_fit_constrained_bad_input_dim():
    """Bad input for constrained"""
    XTX = OnlineGram(10)
    XTY = OnlineGram(5)

    XTX.update(jax.random.uniform(random.generate_key(), shape=(100, 10)))
    XTY.update(jax.random.uniform(random.generate_key(), shape=(100, 5)))

    with pytest.raises(ValueError):
        fit_gram(XTX, XTY, input_dim=7)



"""flax.nn.Module for a principal component regression online learner."""
from typing import Any
from typing import Iterable
from typing import Tuple
from typing import Union

import flax
import jax.numpy as jnp
import numpy as np

from timecast.learners._linear import Linear
from timecast.learners.base import FitMixin
from timecast.learners.base import NewMixin
from timecast.utils import random
from timecast.utils.ar import compute_gram
from timecast.utils.ar import fit_gram
from timecast.utils.pcr import compute_projection


class PCR(NewMixin, FitMixin, flax.nn.Module):
    """PCR online learner"""

    def apply(
        self,
        x: np.ndarray,
        history_len: int,
        projection: np.ndarray,
        output_dim: Union[Tuple[int, ...], int] = 1,
        history: np.ndarray = None,
        loc: Union[np.ndarray, float] = None,
        scale: Union[np.ndarray, float] = None,
    ):
        """
        Todo:
            * AR doesn't take any history
        output_dim (Union[Tuple[int, ...], int]): int or tuple describing
        output shape

        Note:
            * We expect that `x` is one- or two-dimensional
            * We reshape `x` to ensure its first axis is time and its second
              axis is input_features

        Args:
            x (np.ndarray): input data
            history_len (int): length of AR history length
            output_dim (Union[Tuple[int, ...], int]): int or tuple
            describing output shape
            history (np.ndarray, optional): Defaults to None. Optional
            initialization for history
            loc: mean for centering data
            scale: std for normalizing data

        Returns:
            np.ndarray: result
        """
        if jnp.isscalar(x):
            x = jnp.array([[x]])
        if x.ndim == 1:
            x = x.reshape(1, -1)

        self.history = self.state(
            "history", shape=(history_len, x.shape[1]), initializer=flax.nn.initializers.zeros
        )

        if self.is_initializing() and history is not None:
            self.history.value = jnp.vstack((self.history.value, history))[history.shape[0] :]
        elif not self.is_initializing():
            self.history.value = jnp.vstack((self.history.value, x))[x.shape[0] :]

        inputs = self.history.value.reshape(1, -1)

        if loc is not None:
            inputs -= loc

        if scale is not None:
            inputs /= scale

        inputs @= projection

        y = Linear(
            inputs=inputs,
            output_shape=output_dim,
            input_axes=(0, 1),
            bias=True,
            dtype=jnp.float32,
            kernel_init=flax.nn.initializers.zeros,
            bias_init=flax.nn.initializers.zeros,
            precision=None,
            name="Linear",
        )
        return y

    @classmethod
    def fit(
        cls,
        data: Iterable[Tuple[np.ndarray, np.ndarray, Any]],
        input_dim: int,
        history_len: int,
        output_dim: int = 1,
        k: int = None,
        normalize: bool = True,
        alpha: float = 1.0,
        key: jnp.ndarray = None,
        history: np.ndarray = None,
        name: str = "PCR",
        **kwargs
    ) -> flax.nn.Model:
        """Receives data as an iterable of tuples containing input time series,
        true time series

        Todo:
            * We assume input_dim is one-dimensional; we should flatten if not
            * Really intended for passing in timeseries at a time, not
            individual time series observations; is this the right general API?
            * Shape is (1, input_dim); what about mini-batches?

        Notes:
            * Use (1, history_len * input_dim) vectors as features (could
            consider other representations)
            * Ignore true value (i.e., look at features only) (should we
            consider impact of features on dependent variable?)
            * Given a time series of length N and a history of length H,
            construct N - H + 1 windows
            * We could infer input_dim from data, but for now, require
            users to explicitly provide
            * Assumes we get tuples of time series, not individual time series
            observations

        Args:
            data: an iterable of tuples containing input/truth pairs of time
            series plus any auxiliary value
            input_dim: number of feature dimensions in input
            history_len: length of history to consider
            output_dim: number of feature dimensions in output
            k: number of PCA components to keep. Default is min(num_histories,
            normalize: zscore data or not
            input_dim)
            alpha: Parameter to pass to ridge regression for AR fit
            key: random key for jax random
            history: Any history to pass to AR
            name: name for the top-level module
            kwargs: Extra keyword arguments

        Returns:
            flax.nn.Model: initialized model
        """
        XTX, XTY = compute_gram(data, input_dim, output_dim, history_len)
        num_features = input_dim * history_len

        # Compute k
        min_dim = min(num_features, XTX.observations)
        k = min_dim if k is None else min(k, min_dim)

        projection = compute_projection(XTX.matrix(normalize=normalize), k)

        # X: (n, d)
        # XTX.matrix: (d, d)
        # projection: (d, k)
        # X @ projection: (n, k)
        # (X @ projection).T @ (X @ projection): (k, k)
        # projection.T @ X.T @ X @ projection = (k, d) @ (d, n) @ (n, d) @ (d, k)
        kernel, bias = fit_gram(XTX, XTY, normalize=normalize, projection=projection, alpha=alpha)

        loc = XTX.mean if normalize else None
        scale = XTX.std if normalize else None

        model_def = PCR.partial(
            history_len=history_len,
            projection=projection,
            output_dim=output_dim,
            loc=loc,
            scale=scale,
            history=history,
            name=name,
        )

        if key is None:
            key = random.generate_key()

        with flax.nn.stateful() as state:
            _, params = model_def.init_by_shape(key, [(1, input_dim)])

        model = flax.nn.Model(model_def, params)
        model.params["Linear"]["kernel"] = kernel
        model.params["Linear"]["bias"] = bias

        return model, state



"""timecast.utils.pcr"""
import jax.numpy as jnp
import numpy as np


def compute_projection(X: np.ndarray, k: int, center=False) -> np.ndarray:
    """Compute PCA projection"""

    if center:
        X = X - X.mean(axis=0)
        # X /= X.shape[0] - 1

    # Compute SVD
    # X: (steps - history_len + 1, history_len * input_dim) -> (H, d)
    # U: (H, H), but because full_matrices=False, (H, d)
    # S: (min(H, d),)
    # VT: (d, d)
    U, S, VT = jnp.linalg.svd(X, full_matrices=False, compute_uv=True)

    # Get index of top K eigen values
    top_k = (-jnp.square(S)).argsort()[:k]

    # Get projection
    # projection: (d, k)
    projection = VT[top_k].T

    return projection



"""timecast.objectives: testing"""
import jax.numpy as jnp

from timecast.objectives import residual
from timecast.objectives import xboost


def model(x):
    """Dummy ensemble model"""
    return [1, 2, 3, 4, 5]


def loss_fn(true, pred):
    """Dummy loss function"""
    return jnp.square(true - pred).mean()


def test_residual():
    """Test residual objective"""
    loss, y_hat = residual(0, 10, loss_fn, model)

    assert loss == 90 and y_hat == 15


def test_xboost():
    """Test xboost objective"""
    loss, y_hat = xboost(0.0, 10.0, loss_fn, model)

    assert jnp.isclose(loss, 264.83334) and jnp.isclose(y_hat, 3.6666667)



"""timecast.learners._learner: testing"""
from typing import Iterable

import flax
import jax
import jax.numpy as jnp
import numpy as np
import pytest

from timecast.learners import Linear
from timecast.learners._linear import _normalize_axes
from timecast.utils import random


def get_kernel_bias(model):
    """Utility function to get kernel and bias params"""
    return model.params["kernel"], model.params["bias"]


def eval_model(model, state, inputs):
    """Utility function to evaluate a model with inputs"""
    with flax.nn.stateful(state) as state:
        result = model(inputs)
    return result


@pytest.mark.parametrize(
    "axes,ndim,expected", [((0, 1), 3, (0, 1)), ((-1, 0), 3, (2, 0)), ((4, 0), 3, (4, 0))]
)
def test_normalize_axes(axes, ndim, expected):
    """Test utility function for normalizing axes"""
    assert _normalize_axes(axes, ndim) == expected


@pytest.mark.parametrize("output_shape", [(), 1, (1,), (1, 2), (2, 1), (2, 2), (3, 2, 4)])
@pytest.mark.parametrize("input_axes", [(), -1, 1, (1,), (1, 2), (2, 3)])
@pytest.mark.parametrize("batch_axes", [(), 1, (1,), 0, (0,), (0, 1)])
def test_linear(output_shape, input_axes, batch_axes):
    """Test linear

    Todo:
        * Separate tests for exceptional behavior to get rid of duplicate tests
        and sloppy code
    """
    orig_output_shape, orig_input_axes, orig_batch_axes = output_shape, input_axes, batch_axes

    if not isinstance(output_shape, Iterable):
        output_shape = (output_shape,)
    if not isinstance(input_axes, Iterable):
        input_axes = (input_axes,)
    if not isinstance(batch_axes, Iterable):
        batch_axes = (batch_axes,)
    output_shape, input_axes, batch_axes = (
        tuple(output_shape),
        tuple(input_axes),
        tuple(batch_axes),
    )

    shapes = [
        tuple(jax.random.randint(random.get_key(), (len(batch_axes) + len(input_axes),), 0, 100))
    ]

    if len(input_axes) == 0 or len(output_shape) == 0:
        with pytest.raises(IndexError):
            Linear.new(
                shapes=shapes,
                output_shape=output_shape,
                input_axes=input_axes,
                batch_axes=batch_axes,
            )

    elif batch_axes and set(batch_axes) != set(range(jnp.max(batch_axes) + 1)):
        with pytest.raises(ValueError):
            Linear.new(
                shapes=shapes,
                output_shape=output_shape,
                input_axes=input_axes,
                batch_axes=batch_axes,
            )

    elif set(batch_axes) & set(input_axes):
        with pytest.raises(IndexError):
            Linear.new(
                shapes=shapes,
                output_shape=output_shape,
                input_axes=input_axes,
                batch_axes=batch_axes,
            )

    elif jnp.max(batch_axes + input_axes) >= len(shapes[0]):
        with pytest.raises(IndexError):
            Linear.new(
                shapes=shapes,
                output_shape=output_shape,
                input_axes=input_axes,
                batch_axes=batch_axes,
            )
    else:
        model, state = Linear.new(
            shapes=shapes,
            output_shape=orig_output_shape,
            input_axes=orig_input_axes,
            batch_axes=orig_batch_axes,
        )

        inputs = jax.random.uniform(random.generate_key(), shape=shapes[0])
        ndim = inputs.ndim

        input_axes = _normalize_axes(input_axes, ndim)
        batch_axes = _normalize_axes(batch_axes, ndim)

        kernel, bias = get_kernel_bias(model)
        result = eval_model(model, state, inputs)

        expected = (
            jnp.tensordot(
                inputs, kernel, axes=(input_axes, tuple(range(kernel.ndim - len(output_shape))))
            )
            + bias
        )

        np.testing.assert_array_almost_equal(expected, result)



"""timecast.learners"""
from timecast.learners._ar import AR
from timecast.learners._arx import ARX
from timecast.learners._arx_history import ARXHistory
from timecast.learners._combinators import Parallel
from timecast.learners._combinators import Sequential
from timecast.learners._linear import Linear
from timecast.learners._pcr import PCR
from timecast.learners._precomputed import Precomputed
from timecast.learners._predict_constant import PredictConstant
from timecast.learners._predict_last import PredictLast
from timecast.learners._take import Take
from timecast.learners.base import FitMixin
from timecast.learners.base import NewMixin


__all__ = [
    "AR",
    "ARX",
    "ARXHistory",
    "FitMixin",
    "Linear",
    "NewMixin",
    "PCR",
    "Precomputed",
    "PredictConstant",
    "PredictLast",
    "Parallel",
    "Sequential",
    "Take",
]



"""timecast.learners.pcr: testing"""
import flax
import jax
import jax.numpy as jnp
import numpy as np
import pytest

from tests.learners.test_ar import _compute_kernel_bias
from timecast.learners import PCR
from timecast.utils import internalize
from timecast.utils import random
from timecast.utils.ar import historify
from timecast.utils.pcr import compute_projection


def test_pcr_fit_index_error():
    """Test PCR fit with no data"""
    with pytest.raises(IndexError):
        PCR.fit([], input_dim=1, history_len=1)


@pytest.mark.parametrize("shape", [(100, 1), (200, 5)])
@pytest.mark.parametrize("history_len", [1, 2, 5])
def test_pcr_fit(shape, history_len):
    """Test PCR fit"""
    X = jax.random.uniform(random.generate_key(), shape=shape)
    Y = jax.random.uniform(random.generate_key(), shape=(shape[0],))

    pcr, state = PCR.fit(
        [(X, Y, None)], input_dim=shape[1], history_len=history_len, normalize=False, alpha=1.0
    )

    X = internalize(X, shape[1])[0]
    Y = internalize(Y, 1)[0]

    num_histories = X.shape[0] - history_len + 1
    X = historify(X, num_histories=num_histories, history_len=history_len, flatten=True)
    Y = Y[-len(X) :]

    k = shape[1] * history_len

    projection = compute_projection(X, k)
    kernel, bias = _compute_kernel_bias(X @ projection, Y, alpha=1.0)
    kernel = kernel.reshape(1, k, 1)

    np.testing.assert_array_almost_equal(
        abs(kernel), abs(pcr.params["Linear"]["kernel"]), decimal=3
    )
    np.testing.assert_array_almost_equal(abs(bias), abs(pcr.params["Linear"]["bias"]), decimal=3)


def test_pcr_apply():
    """Test PCR apply shapes"""
    pcr, state = PCR.new(
        shapes=[(1, 1)],
        history_len=2,
        projection=jnp.eye(2),
        loc=0,
        scale=1,
        history=jnp.ones((2, 1)),
    )

    with flax.nn.stateful(state) as state:
        scalar = pcr(1)

    with flax.nn.stateful(state) as state:
        vector = pcr(jnp.ones((1,)))

    np.testing.assert_array_almost_equal(scalar, vector)


def test_pcr_fit_value_error():
    """Test number of observations"""
    X = jax.random.uniform(random.generate_key(), shape=(1, 10))
    Y = jax.random.uniform(random.generate_key(), shape=(10, 1))

    with pytest.raises(ValueError):
        PCR.fit([(X, Y, None)], input_dim=10, history_len=1)
